(* Daan leijen 2001, daan@cs.uu.nl *)
open Optspace
open Optnorm
open Optlocal

let tickcount ()
  = Sys.time ()

(* optimize goes into a fixpoint loop. 
  time   - the starting time of the program
  time0  - the time at which "optimize" is called first
  tlimit - the time limit
  pass   - the current pass
  fix    - if true, the previous pass didn't compress the tree.
           we always try this twice, since a second pass might compress again.
  size0  - the size of the current tree
  ts0    - the current tree
*)
let rec optimize time time0 tlimit pass fix size0 ts0
  = begin
      let ts1   = opt_norm (opt_local ts0) in
      let time1 = tickcount () in
      if (time1 > time +. tlimit)
       then ts1
       else let size1 = Sml.size_of_trees ts1 in
            if (size1 > size0)      (* stupidity? *)
             then ts0
             else let time_avg  = ((time1 -. time0) /. float_of_int pass) in
                  if (time1 +. time_avg < time +. tlimit) (* reasonable amount of time left? *)
                   then if (size1 = size0) (* fixpoint ? *)
                         then (if (fix) then ts1
                                        else optimize time time0 tlimit (pass+1) true size1 ts1)
                         else optimize time time0 tlimit (pass+1) false size1 ts1
                   else ts1 (* no time left *)
    end

let main
  = begin
      let time    = tickcount () in
      let tlimit  = try float_of_string (Sys.getenv "TLIMIT") with _ -> (24.0*.60.0*.60.0) in      
      let tlimit0 = tlimit *. 0.9 in  (* take a safe threshold *)
      
      let ts0    = opt_norm (opt_space (Parser.parse ())) in
      let size0  = Sml.size_of_trees ts0 in
      let time0  = tickcount () in
      let ts1    = if (time0 < time +. tlimit0)
                    then optimize time time0 tlimit0 1 false size0 ts0 
                    else ts0 in
      Print.print stdout ts1
    end


