(* Daan leijen 2001, daan@cs.uu.nl *)
open Sml

let insert_tag tag tags
  = match tag with
    | Size sz  -> tag::List.filter (fun t -> not (is_size t)) tags
    | Color c  -> tag::List.filter (fun t -> not (is_color t)) tags
    | S        -> S::List.filter (fun t -> (t<>EM && t<>S)) tags
    | EM       -> if (List.mem EM tags)
                   then List.filter (fun t -> t<>EM) tags
                  else if (List.mem S tags)
                   then tags
                   else EM::tags
    | PL       -> PL::List.filter (fun t -> not (is_pl_sensitive t)) tags
    | other    -> if (List.mem tag tags)
                   then tags
                   else tag::tags

let merge_tag (tags0,tags1) tag
  = match tag with
    | B  when not (List.mem PL tags0) -> (insert_tag B tags0,tags1)
    | I  when not (List.mem PL tags0) -> (insert_tag I tags0,tags1)
    | TT when not (List.mem PL tags0) -> (insert_tag TT tags0,tags1)

    | S  when not (List.exists (fun t -> (t=PL)) tags0) (* allow EM, cancels out in insert_tag *)
          -> (insert_tag S tags0,tags1)
    | EM when not (List.exists (fun t -> (t=PL)) tags0)  (* allow S and EM, cancels out in insert_tag *)
          -> (insert_tag EM tags0,tags1)
    | PL  -> (insert_tag PL tags0,tags1)
    
    | Size sz -> (insert_tag tag tags0,tags1)
    | Color c -> (insert_tag tag tags0,tags1)
    | other   -> (tags0,insert_tag tag tags1)



let merge_tags tags0 tags1
  = if (null tags0)
     then (tags1,[])
     else let x = List.fold_left merge_tag (tags0,[]) tags1 in
          assert(next_ctxs root_ctx (tags0 @ tags1) = next_ctxs root_ctx (fst x @ snd x));
          x

(* collapse 'threads' of tags that don't influence each other *)
let rec opt_thread tags0 t
  = match t with
    | Text txt  
        -> if (null tags0) 
            then t
           else if (String.length txt = 0)
            then Tag ([],[])    (* tags do nothing *)
            else Tag (tags0,[t])
    | Tag (tags,[])
        -> Tag ([],[]) (* tags do nothing *)
    | Tag (tags,[t])
        -> let (pre,post) = merge_tags tags0 tags
           in if (null post)
               then opt_thread pre t
               else Tag (pre,[opt_thread post t])
    | Tag (tags,body)
        -> let (pre,post) = merge_tags tags0 tags
           in if (null post)
               then Tag (pre,opt_body body)
               else Tag (pre,[Tag (post,opt_body body)])

and opt_body ts
  = List.map (opt_thread []) ts

(* normalize SML trees, ie. collect cascading tags *)
let opt_norm ts
  = opt_body ts


