(* Daan Leijen 2001, daan@cs.uu.nl *)
open Sml
open Parser

(* implement a lookahead of 1 character *)
let eof
  = char_of_int 0

let lookahead
  = ref eof

let prim_next_char ()
  = try let c  = input_char stdin in
        if (c == char_of_int 4) then eof else c
    with End_of_file -> eof

let init_lookahead ()
  = lookahead := prim_next_char ()

let next_char ()
  = begin
      let c = !lookahead in
      lookahead := prim_next_char ();
      c
    end

let peek_char ()
  = !lookahead

let rec skip_till c0
  = let c1 = next_char () in
    if (c0 == c1 || c1 == eof)
     then ()
     else skip_till c0
    
(* the parser, assumes legal input! *)
let rec parse_tag ()
  = let context tag 
          = begin
              skip_till '>';
              let ts = parse_document () in
              skip_till '>';
              Tag ([tag],ts)
            end in
    match next_char () with
    | 'B'   -> context B
    | 'E'   -> context EM
    | 'I'   -> context I
    | 'P'   -> context PL
    | 'S'   -> context S
    | 'T'   -> context TT
    | 'U'   -> context U
    | c     -> if (c >= '0' && c <= '9')
                then context (Size (int_of_char c - int_of_char '0'))
                else context (Color c)

and parse_text len acc
  = let c = peek_char () in
    if (c == '<' || c == eof)
     then Text (string_of_revlist_len len acc)
     else let c = next_char () in
          parse_text (len+1) (c::acc)
  

and parse_document ()
  = let c = next_char () in
    if (c == '<')
     then if (peek_char () != '/')
           then let t  = parse_tag () in
                let ts = parse_document () in
                t::ts
           else []
    else if (c == eof)
     then []
     else let t  = parse_text 1 [c] in
          let ts = parse_document () in
          t::ts

(* read an SML tree from stdin *)
let parse ()
  = begin
      init_lookahead ();
      parse_document ()
    end
